/*
 * Autor: Marcin Szewczyk, 198370
 *
 * Temat: smart_ptr
 *
 * Inteligentne wskaniki zawarte w module smart_ptr biblioteki boost.
 * Sposb ich wykorzystania, rnice pomidzy poszczeglnymi typami wskanikw, 
 * nieco ciekawostek, ktre nie rzucaj si od razu w oczy.
 *
 * Chocia ten plik jest do dugi zawiera tylko gar informacji. Pena 
 * dokumentacja, na ktrej opieraem si piszc ten program jest tu:
 * http://www.boost.org/doc/libs/1_35_0/libs/smart_ptr/smart_ptr.htm
 *
 * Omwione tutaj typy inteligentnych wskanikw to:
 *  - scoped_ptr
 *  - scoped_array
 *  - shared_ptr
 *  - weak_ptr
 *  - intrusive_ptr
 *  - oraz std::auto_ptr nienalecy do bibl. boost, ale warty porwnania
 *
 *  nie zosta opisany typ shared_array, poniewa nie wnosi nic nowego do 
 *  omawianych tu zagadnie
 *
 *  Changelog:
 *  2008.10.18  W kocu poprawiem may wyciek pamici w miejscu, w ktrym
 *              nie uyem smart_ptr - o ironio!
 *
 *  Dobre rady:
 *  W tym dokumencie prezentowany jest wyciek pamici w przypadku cykli
 *  budowanych na shared_ptr i jego brak w przypadku cykli budowanych na
 *  weak_ptr. Mona uzyska dobr wizualizacj zagadnienia przy pomocy
 *  aplikacji valgrind,
 *  Proponowany sposb uycia:
 *  valgrind --leak-check=full ./smart_ptr > /dev/null
 *  Efekt:
 *      114 (12 direct, 102 indirect) bytes in 1 blocks are definitely lost in loss record 3 of 5
 *        at 0x4023294: operator new(unsigned) (vg_replace_malloc.c:224)
 *        by 0x8049AAE: shared_cycle() (smart_ptr.cpp:474)
 *        by 0x804B5C6: main (smart_ptr.cpp:608)
 *
 *  Support:
 *      wodny@thlen.pl <- usu h
 */

#include <cstdio>
#include <iostream>
#include <cstring>
#include <boost/scoped_ptr.hpp>
#include <boost/scoped_array.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <boost/intrusive_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <vector>
#include <exception>

// ==========================================================
// Klasy potrzebne do prezentacji podstawowych cech inteligentnych wskanikw

class Harry {
	public:
		Harry(const char* = NULL);
		~Harry();

		void hello();
	private:
		char *name;
};


class Betty : public Harry {
	public:
		Betty();
		~Betty();
};

// ----------------------------------------------------------

Harry::Harry(const char *name_) {
	name = name_ ? strdup(name_) : NULL;
	printf("-- Constructing Harry (%s)\n", name);
}

Harry::~Harry() {
	printf("-- Deconstructing Harry (%s)\n", name);
        free(name);
}

void Harry::hello() {
	printf("-- Harry \"%s\" says hello\n", name);
}

// ----------------------------------------------------------

Betty::Betty() {
	printf("---- Constructing Betty\n");
}

Betty::~Betty() {
	printf("---- Deconstructing Betty\n");
}

// ==========================================================
// Klasa, ktra posuy do prezentacji, jak radz sobie inteligentne wskaniki z 
// cyklami
//
// Wariant dla cyklu z uyciem shared_ptr

class RingNode {
	public:
		RingNode(const char*, const boost::shared_ptr<RingNode>&);
		~RingNode();
		void connect(const boost::shared_ptr<RingNode>&);
		const char* get_name();
		const boost::shared_ptr<RingNode>& get_node();
	private:
		boost::shared_ptr<RingNode> node;
		char* name;
};

RingNode::RingNode(const char* name_, const boost::shared_ptr<RingNode>& node_) : node(node_) {
	name = name_ ? strdup(name_) : NULL;
	printf("++ Constructing RingNode (%s)\n", name);
}

RingNode::~RingNode() {
	printf("++ Deconstructing RingNode (%s)\n", name);
	free(name);
}

void RingNode::connect(const boost::shared_ptr<RingNode>& node_) {
	node = node_;
}

const char* RingNode::get_name() {
	return name;
}

const boost::shared_ptr<RingNode>& RingNode::get_node() {
	return node;
}

// ----------------------------------------------------------
// Wariant cyklu z weak_ptr

class WeakRingNode {
	public:
		WeakRingNode(const char*, const boost::shared_ptr<WeakRingNode>&);
		~WeakRingNode();
		void connect(const boost::shared_ptr<WeakRingNode>&);
		const char* get_name();
		const boost::shared_ptr<WeakRingNode> get_node();
	private:
		boost::weak_ptr<WeakRingNode> node;
		char* name;
};

WeakRingNode::WeakRingNode(const char* name_, const boost::shared_ptr<WeakRingNode>& node_) : node(node_) {
	name = name_ ? strdup(name_) : NULL;
	printf("++ Constructing RingNode (%s)\n", name);
}

WeakRingNode::~WeakRingNode() {
	printf("++ Deconstructing RingNode (%s)\n", name);
        free(name);
}

void WeakRingNode::connect(const boost::shared_ptr<WeakRingNode>& node_) {
	node = node_;
}

const char* WeakRingNode::get_name() {
	return name;
}

const boost::shared_ptr<WeakRingNode> WeakRingNode::get_node() {
	return node.lock();
}

// ==========================================================
// Klasa potrzebna do prezentacji intrusive_ptr
//  czyli sama zwiksza/zmniejsza ilo referencji do danego obiektu na zlecenie 
//  intrusive_ptr oraz sama odpowiada za zniszczenie obiektu, gdy ilo odwoa 
//  spadnie do zera

class Intruded {
	public:
		Intruded();
		~Intruded();
	private:
		int counter;

		friend void intrusive_ptr_add_ref(Intruded *const);
		friend void intrusive_ptr_release(Intruded *const);
};

void intrusive_ptr_add_ref(Intruded *const);
void intrusive_ptr_release(Intruded *const);

// ----------------------------------------------------------

Intruded::Intruded() : counter(0) {
	printf("%% Creating Intruded\n");
}

Intruded::~Intruded() {
	printf("%% Destroying Intruded\n");
}

// Jak wida, licznik jest skadow klasy, natomiast funkcje odpowiedzialne za 
// zliczanie - nie.
// Na podstawie typu parametru, intrusive pointer wybiera odpowiedni zestaw 
// funkcji do inkrementacji i dekrementacji iloci odwoa

void intrusive_ptr_add_ref(Intruded *const p) {
	printf("%% Added. Now: %d\n", ++(p->counter));
}

void intrusive_ptr_release(Intruded *const p) {
	printf("%% Removed. Now: %d\n", --(p->counter));
	if(!p->counter) {
		printf("%% Deleting Intruded because of count == 0...\n");
		delete p;
	}
}

// ==========================================================
// Jak uzyska poprawny (pod wzgldem zliczania odwoa) shared_ptr do obiektu 
// nie majc wiedzy na temat tego, czy ju jaki inny shared_ptr si do niego 
// odwouje

// Konieczne jest dziedziczenie klasy po enable_shared_from_this<>

class WithSharing : public boost::enable_shared_from_this<WithSharing> {
public:
	boost::shared_ptr<WithSharing> get_shared() {
		return shared_from_this();
	}
};

// ==========================================================
// ==========================================================
// ==========================================================

using namespace std;

// By unikn bardzo dugich definicji poniej
typedef vector< boost::shared_ptr<Harry> > SharedHarryVector;

// ==========================================================
// Funkcja wywietla nagwek tekstowy podkrelony linijk

void header(const char* label) {
	char* bar = strdup(label);
	cout << endl << label << endl;
	memset(bar, '=', strlen(bar));
	cout << bar << endl;
        free(bar);
}

// ==========================================================
// ==========================================================

// Funkcja jest wywoywana, ale szkodliwy fragment jest wykomentowany
void play_dirty_harry() {
	header("Dirty Harry - czego ze wskaznikami robic nie wolno");

	Harry* tmp_harry;
	boost::scoped_ptr<Harry> harry(tmp_harry = new Harry("Simple Class Harry"));

	//boost::scoped_ptr<Harry> harry2(tmp_harry);
	// Nielegalne - dwa scoped_ptr, kady z liczb odwoa = 1
	// Podczas destrukcji dwch scoped_ptr dochodzi dwa razy do prby destrukcji 
	// jednego obiektu klasy Harry. Lepiej przedstawi to shared_ptr, ktry pojawi 
	// si poniej

	//boost::scoped_ptr<Harry> harry_b(harry);
	// Kopiowanie scoped_ptr jest zakazane (konst. kopiujcy w sekcji private)
	//boost::scoped_ptr<Harry> harry_c;
	//harry_c = harry;
	// Analogiczna uwaga dotyczy operatora przypisania

}

void play_nice_harry() {
	header("play_nice_harry - scoped_ptr jak zmienna automatyczna");
	// Jakie tam zadania
	// ...

	// Tutaj proste zastosowanie - scoped_ptr zachowuje si jak zmienna 
	// automatyczna. Po wyjciu poza zasig, scoped_ptr niszczy siebie, a 
	// wczeniej zwizany z nim obiekt.
	//
	// Std nazwa - scoped - czyli dot. zasigu.
	// Nawizuje to do pomysu "resource acquisition is initialization". Zasoby 
	// s zajmowane w momencie inicjalizacji i, co waniejsze, zwalniane podczas 
	// "odinicjowania", czyli destrukcji po wyjciu poza zasig.

	boost::scoped_ptr<Harry> harry3(new Harry("Harry3"));
	harry3->hello();

	Harry harry4("Harry4");
	harry4.hello();

}

void play_nice_harry3() {
	// Jednak nie zawsze wystarcza nam zmienna automatyczna. Ciekawsze jest 
	// dynamiczne powoywanie obiektw do ycia operatorem "new".

	header("play_nice_harry3 - operator \"new\" bez potrzeby wywolania delete");
	boost::scoped_ptr<Harry> harry7;

	//
	// Co tam do roboty...
	bool needHarry = true;
	// Co tam do roboty...
	//

	if(needHarry)
		harry7.reset(new Harry("Harry7"));

	//
	// Co tam do roboty...
	//

	if(harry7) {
		printf("-- Stwierdzlismy, ze harry7 sie przyda - wykonajmy na nim operacje\n");
		// Operacje na Harrym...
	}

	// Normalnie w tym miejscu przydaoby si wywoanie delete na zwykym 
	// wskaniku. Poniewa jednak nasz inteligenty wskanik jest zmienn 
	// automatyczn, sam usunie siebie i skojarzony obiekt
}

void comfort_test() throw(exception) {
	header("Comfort test - shared_ptr jak zm. auto.; auto. usuwanie, gdy rzucany jest wyjatek");
	boost::shared_ptr<Harry> h1;
	boost::shared_ptr<Harry> h2;

	// Co tam...

	// Potrzebujemy Harry'ego
	h1.reset(new Harry("Potter"));
	h2.reset(new Harry("Callahan"));
	cout << "Callahan count: " << h2.use_count() << endl;
	h1 = h2;
	cout << "Callahan zjadl Pottera, count: " << h2.use_count() << endl;
	cout << "h1: ";
		h1->hello();
	cout << "h2: ";
		h2->hello();

	// Go ahead, make my day.
	//   Rzucamy wyjtek
	cout << "Throwing an exception .44" << endl;
	throw exception();
}

// Prezentacja, jak zachowuj si intrusive_ptr
void test_intrusive_ptr() {
	header("test_intrusive_ptr");
	boost::intrusive_ptr<Intruded> i1(new Intruded);
}

// Ciekawostka zwizana z shared_ptr
// Rnica, midzy tym, co deklarujemy jako klas dla szablonu shared_ptr, a 
// klas parametru przekazywanego w konstruktorze. Jakie nieoczywiste zdarzenie 
// moe wystpi, gdy destruktory klas s niewirtualne.

void non_virt_diffs() {
	header("Niewirtualny destruktor. Roznice miedzy zwyklym wskaznikiem a shared_ptr");
	// Destructor nie jest wirtualny, std Betty nie jest niszczona
	{
		cout << "Zwykly Harry-wskaznik na Betty" << endl;
		Harry *hb = new Betty;
		delete hb;
	}
	
	// Nieco inaczej sprawa si przestawia dla shared_ptr
	{
		cout << endl << "Betty-wskaznik na Betty w Harry-szablonie" << endl;
		boost::shared_ptr<Harry> b1(new Betty);
		// shared_ptr pamita rzeczywisty typ wskanika przekazywanego w 
		// parametrze. W zwizku z tym uywa destruktora dla tego typu (Betty) 
		// zamiast typu deklarowanego w szablonie (Harry)
		//  > przykadowa implementacja mechanizmu:
		//  > http://wodny.org/download/smart_ptr/template_memory.cpp
	}
	
	{
		cout << endl << "Harry-wskaznik na Betty w Harry-szablonie" << endl;
		Harry *b3 = new Betty;
		boost::shared_ptr<Harry> b2(b3);
		// Tutaj shared_ptr nie ma okazji pozna prawdziwego typu obiektu.  
		// Traktuje go jak Harry'ego. Dla poprawnego dziaania, tu destruktor ju 
		// musi by wirtualny.
	}

}

void play_scoped_array() {
	// Prosty przykad, e inteligentne wskaniki z dopiskiem _array su do 
	// kojarzenia ich z tablicami. Nie mona by tu uy scoped_ptr. Jest to 
	// zwizane z obostrzeniem, e delete stosujemy dla pojedynczego obiektu, 
	// delete[] dla tablic. Inteligenty wskanik nie posiadajc informacji o 
	// naturze przekazanego wskanika le zarzdzaby pamici.
	header("Zabawy ze scoped_array");
	boost::scoped_array<Harry> harry_a1(new Harry[3]);
}

void general_use() {
	header("Ogolny sposob uzycia shared_ptr");

	Harry *hp1 = NULL;
	boost::shared_ptr<Harry> harry_sp1(hp1 = new Harry("Shared Harry1"));
	cout << "harry_sp1.use_count: " << harry_sp1.use_count() << endl;
	//boost::shared_ptr<Harry> harry_sp2(hp1);
	// Taka konstrukcja spowoduje bd. Dwa shared_ptr bd prboway usun dwa 
	// razy ten sam obiekt podczas wasnej destrukcji.
	//
	// Istotne jest, by zauway, e shared_ptr nie ma jakiej ukrytej tablicy 
	// asocjacyjnej (trzymajcej si np. wzorca Singleton), ktra pozwalaaby mu 
	// stwierdza, ktrymi wskanikami ju si opiekuje.
	// Poinformowanie o tym, e kolejny shared_ptr chce rwnie odnosi si do 
	// danego zwykego wskanika nastpuje przez wykonanie konstruktora 
	// kopiujcego (opcjonalnie operatora przypisania). Od tej pory kady z 
	// shared_ptr ustawi ilo odwoa na 2 i przy destrukcji jednego z nich, 
	// drugi inteligenty wskanik dowie si o tym zdarzeniu i zdekrementuje 
	// wasny licznik odwoa
	// (Mowa jest tu o logice dziaania, nie rzeczywistej implementacji 
	// inteligentnego wskanika).
	cout << "Czas, by podzielic sie zasobem" << endl;
	boost::shared_ptr<Harry> harry_sp2(harry_sp1);
	boost::shared_ptr<Harry> harry_sp3;
	harry_sp3 = harry_sp2;
	cout << "harry_sp1.use_count: " << harry_sp1.use_count() << endl;
	cout << "harry_sp2.use_count: " << harry_sp2.use_count() << endl;
	cout << "harry_sp3.use_count: " << harry_sp3.use_count() << endl;

}

SharedHarryVector* get_harries() {
	// Shared_ptr a kontenery
	// (scoped_ptr nie moe zosta uyty do takich zastosowa - nie da si 
	// skopiowa kontenerowi, poza tym jego przeznaczenie jest inne)
	header("Wektor (vector) z Harrych - uzycie shared_ptr w kontenerach STL");

	SharedHarryVector *vect = new SharedHarryVector();
	boost::shared_ptr<Harry> h;
	for(int i = 0; i < 4; ++i) {
		h.reset(new Harry("Vector Harry"));
		cout << "Ilosc Harry'ego przed wlozeniem do wektora: " << h.use_count() << endl;
		vect->push_back(h);
		cout << "Po wlozeniu: " << h.use_count() << endl;
	}
	return vect;
}

void stl_container() {
	// W nawizaniu do powyszej funkcji
	boost::shared_ptr< SharedHarryVector > harry_vector(get_harries());
	cout << "Liczno jednego z wektorowych Harrych: " << harry_vector->front().use_count() << endl << endl;
}

void no_cycle() {
	header("Bez cyklu");
	// Bez cyklu
	boost::shared_ptr<RingNode> R3(new RingNode("RingNode3", boost::shared_ptr<RingNode>((RingNode*)NULL)));
	boost::shared_ptr<RingNode> R4(new RingNode("RingNode4", boost::shared_ptr<RingNode>(R3)));
	cout << "R3 (" << R4->get_name() << ") with count " << R3.use_count() << endl;
	cout << "R4 (" << R4->get_name() << ") with count " << R4.use_count() << " connected to " << R4->get_node()->get_name() << endl;
	// Wszystkie wzy ulegn zniszczeniu jak powinny
}

void shared_cycle() {
	header("Cykl na shared_ptr");
	// Wikszy cykl, cho uzyskany tylko porednio, powoduje, e wzy nie 
	// zostan usunite, poniewa ilo odwoa w shared_ptr nie spadnie do zera
	boost::shared_ptr<RingNode> R7(new RingNode("RingNode7", boost::shared_ptr<RingNode>((RingNode*)NULL)));
	boost::shared_ptr<RingNode> R8(new RingNode("RingNode8", boost::shared_ptr<RingNode>(R7)));
	boost::shared_ptr<RingNode> R9(new RingNode("RingNode9", boost::shared_ptr<RingNode>(R8)));
	R7->connect(R9);
	cout << "R7 (" << R7->get_name() << ") with count " << R7.use_count() << " connected to " << R7->get_node()->get_name() << endl;
	cout << "R8 (" << R8->get_name() << ") with count " << R8.use_count() << " connected to " << R8->get_node()->get_name() << endl;
	cout << "R9 (" << R9->get_name() << ") with count " << R9.use_count() << " connected to " << R9->get_node()->get_name() << endl;

	/*
    R7-->"R7"<--------node--------.
           |                      |
           *--node-->"R8"<--R8    |
                       |          |
                       *--node-->"R9"<--R9
	*/
}

void weak_cycle() {
	header("Cykl na weak_ptr");
	// "Saby" cykl oparty o weak_ptr
	boost::shared_ptr<WeakRingNode> R5(new WeakRingNode("WeakRingNode5", boost::shared_ptr<WeakRingNode>((WeakRingNode*)NULL)));
	boost::shared_ptr<WeakRingNode> R6(new WeakRingNode("WeakRingNode6", boost::shared_ptr<WeakRingNode>(R5)));
	R5->connect(R6);
	boost::weak_ptr<WeakRingNode> Wtmp1(R5);
	boost::weak_ptr<WeakRingNode> Wtmp2(R6);
	// Jak mona zauway, weak_ptr nie zwiksza licznika odwoa do obiektu, 
	// ktrego dotyczy, ani nie usuwa obiektu, gdy wielko ta spadnie do zera.  
	// Obiekt, do ktrego si odwouje moe zosta usunity "bez jego wiedzy".  
	// Dlatego wanie wsppracuje z shared_ptr.
	cout << "R5 (" << R5->get_name() << ") with count " << R5.use_count() << " connected to " << R5->get_node()->get_name() << endl;
	cout << "R6 (" << R6->get_name() << ") with count " << R6.use_count() << " connected to " << R6->get_node()->get_name() << endl;
	// Poprzez funkcj weak_ptr.lock(), ktra uyta jest w get_node(), weak_ptr 
	// gwarantuje nam, e zostanie wykonana dodatkowa kopia, a shared_ptr zajmie 
	// si zwikszeniem licznika odwoa, by obiekt nie znikn podczas 
	// wykonywania na nim operacji.
	// Jeli obiekt zdy znikn ju wczeniej, podczas tworzenia shared_ptr z 
	// weak_ptr dostaniemy wyjtek bad_weak_ptr.
}

void shared_and_weak() {
	header("Relacja miedzy shared_ptr i weak_ptr");

	cout << "Oryginalny shared_ptr:" << endl;
	boost::shared_ptr<Harry> shared1(new Harry);
	cout << "shared1.count: " << shared1.use_count() << endl;

	cout << "weak_ptr z shared_ptr:" << endl;
	boost::weak_ptr<Harry> weak1(shared1);
	cout << "shared1.count: " << shared1.use_count() << endl;
	cout << "weak1.count  : " << weak1.use_count() << endl;

	cout << "shared_ptr z weak_ptr (z shared_ptr)" << endl;
	boost::shared_ptr<Harry> shared2(weak1);
	cout << "shared1.count: " << shared1.use_count() << endl;
	cout << "weak1.count  : " << weak1.use_count() << endl;
	cout << "shared2.count: " << shared2.use_count() << endl;
}

void with_sharing_test(WithSharing& ws) {
	header("Test enable_shared_from_this<>");

	boost::shared_ptr<WithSharing> ws_a(ws.get_shared());
	cout << "Drugi udzial, use_count: " << ws_a.use_count() << endl;
	/*  .
	 * /|\
	 *  |
	 * -------------------------------------------------
	 * dwa niezalene fragmenty kodu, np. dwa wtki
	 * -------------------------------------------------
	 *  |
	 * \|/
	 *  '
	 */
	boost::shared_ptr<WithSharing> ws_b(ws.get_shared());
	cout << "Trzeci udzial, use_count: " << ws_b.use_count() << endl;

}

void auto_ptr_demo() {
	header("auto_ptr - przy boost maly sens stosowania");
	/*
	 * http://gcc.gnu.org/onlinedocs/libstdc++/libstdc++-html-USERS-4.0/classstd_1_1auto__ptr.html
	 *
	 * "auto_ptr does not meet the CopyConstructible and Assignable
	 * requirements for Standard Library container elements and thus
	 * instantiating a Standard Library container with an auto_ptr results in 
	 * undefined behavior."
	 *
	 * Reasumujc - auto_ptr ciekawy nie jest. Potrafi tylko oddawa powizanie z 
	 * obiektem, natomiast nie potrafi si nim dzieli. Zastosowanie w 
	 * kontenerach nie jest moliwe.
	 */

	auto_ptr<Harry> a1(new Harry);
	cout << "NEW. " << a1.get() << endl;
	auto_ptr<Harry> a2(a1);
	cout << "1st transfer, source: " << a1.get() << endl;
	cout << "1st transfer, dest. : " << a2.get() << endl;
	auto_ptr<Harry> a3 = a2;
	cout << "2nd transfer, source: " << a2.get() << endl;
	cout << "2nd transfer, dest. : " << a3.get() << endl;
}

// ==========================================================
// ==========================================================
// ==========================================================

int main(int argc, char** argv) {
	// ==========================================================
	// Zestaw tego, czego robi nie wolno ze scoped_ptr
	play_dirty_harry();

	// shared_ptr
	general_use();

	// Mniej interesujce
	play_nice_harry();
	// Bardziej interesujce
	play_nice_harry3();
	
	// scoped_array zamiast scoped_ptr - waciwe delete, konkretnie delete[]
	play_scoped_array();

	// Niewirtualny destruktor. Rnice midzy zwykym wskanikiem a shared_ptr
	non_virt_diffs();

	// ==========================================================
	// Kontener STL
	stl_container();

	// ==========================================================
	// Shared vs Weak a cykle

	no_cycle();
	shared_cycle();
	weak_cycle();
	shared_and_weak();

	// ==========================================================
	//  GWNY POWD UYWANIA SMART_PTRw
	//    komfort operatora "new"
	//    z bezpieczestwem zmiennych automatycznych
	// ==========================================================
	try {
		comfort_test();
	} catch (...) {
		cout << "Caught!" << endl;
	}

	// ==========================================================
	// Test enable_shared_from_this<>
	boost::shared_ptr<WithSharing> ws(new WithSharing);
	with_sharing_test(*ws);

	// ==========================================================
	// std::auto_ptr
	auto_ptr_demo();

	// ==========================================================
	// Intrusive test
	test_intrusive_ptr();

	return 0;
}


